#define _GNU_SOURCE

#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <pwd.h>
#include <grp.h>
#include <sys/fsuid.h>
#include <sys/errno.h>
#include <linux/capability.h>
#include <linux/prctl.h>

#include "giraffe/girtypes.h"
#include "giraffe/mallocvar.h"
#include "giraffe/copyfile.h"
#include "shhopt/shhopt.h"

#include "gircap.h"

/* prctl() is in glibc 2.0, but there is no header file for it.
   It actually takes a variable number of arguments, the number
   determined by the first ("option") argument.
*/

long prctl(int, int);

/* setresuid() and setresgid() are in glibc 2.0, but there is no header
   file for them.
*/
int setresuid(uid_t const ruid,
              uid_t const euid,
              uid_t const suid);
int setresgid(gid_t const rgid,
              gid_t const egid,
              gid_t const sgid);

#define ARRAYSIZE(array) (sizeof(array) / sizeof(array[0]))

struct uidSet {
    /* Each is -1 to mean "same as before" */
    uid_t ruid;
    uid_t euid;
    uid_t suid;
    uid_t fsuid;
};

struct gidSet {
    /* Each is -1 to mean "same as before" */
    gid_t rgid;
    gid_t egid;
    gid_t sgid;
    gid_t fsgid;
};



struct cmdlineInfo {
    /* All the information the user supplied in the command line,
       in a form easy for the program to use.
    */
    unsigned int  setCapabilities;
    capset_t      capabilities;
    struct uidSet uidSet;
    struct gidSet gidSet;
    const char *  execProgName;
    const char ** mainArgs;
    const char *  arg0;
    unsigned int  nogroups;
    unsigned int  asnobody;
    const char *  username;  // NULL if none
};



static void
makeNobodyIds(struct uidSet * const uidSetP,
              struct gidSet * const gidSetP) {

    uid_t const nobodyUid = 999;
    gid_t const nobodyGid = 999;
    
    uidSetP->ruid   = nobodyUid;
    uidSetP->euid   = nobodyUid;
    uidSetP->suid   = nobodyUid;
    uidSetP->fsuid  = nobodyUid;

    gidSetP->rgid   = nobodyGid;
    gidSetP->egid   = nobodyGid;
    gidSetP->sgid   = nobodyGid;
    gidSetP->fsgid  = nobodyGid;
}



static void
parseCommandLine(int                  const argc, 
                 char **              const argv,
                 struct cmdlineInfo * const cmdlineP,
                 const char **        const errorP) {
/*----------------------------------------------------------------------------
   Parse program command line described in Unix standard form by argc
   and argv.  Return the information in the options as *cmdlineP.  

   If command line is internally inconsistent (invalid options, etc.),
   issue error message to stderr and abort program.

   Note that the strings we return are stored in the storage that
   was passed to us as the argv array.  We also trash *argv.
-----------------------------------------------------------------------------*/
    optEntry *option_def;
        /* Instructions to optParseOptions3 on how to parse our options.
         */
    optStruct3 opt;
    int argc2;

    unsigned int option_def_index;

    char * capabilities;
    unsigned int arg0Spec;
    unsigned int ruidSpec, euidSpec, suidSpec, fsuidSpec;
    unsigned int rgidSpec, egidSpec, sgidSpec, fsgidSpec;
    unsigned int usernameSpec;

    unsigned int specificUidOpts, specificGidOpts, specificIdOpts;


    MALLOCARRAY_NOFAIL(option_def, 100);

    option_def_index = 0;   /* incremented by OPTENT3 */
    OPTENT3(0, "capabilities", OPT_STRING, &capabilities,
            &cmdlineP->setCapabilities,  0);
    OPTENT3(0, "ruid",         OPT_UINT,   &cmdlineP->uidSet.ruid,
            &ruidSpec,                   0);
    OPTENT3(0, "euid",         OPT_UINT,   &cmdlineP->uidSet.euid,
            &euidSpec,                   0);
    OPTENT3(0, "suid",         OPT_UINT,   &cmdlineP->uidSet.suid,
            &suidSpec,                   0);
    OPTENT3(0, "fsuid",        OPT_UINT,   &cmdlineP->uidSet.fsuid,
            &fsuidSpec,                  0);
    OPTENT3(0, "rgid",         OPT_UINT,   &cmdlineP->gidSet.rgid,
            &rgidSpec,                   0);
    OPTENT3(0, "egid",         OPT_UINT,   &cmdlineP->gidSet.egid,
            &egidSpec,                   0);
    OPTENT3(0, "sgid",         OPT_UINT,   &cmdlineP->gidSet.sgid,
            &sgidSpec,                   0);
    OPTENT3(0, "fsgid",        OPT_UINT,   &cmdlineP->gidSet.fsgid,
            &fsgidSpec,                  0);
    OPTENT3(0, "arg0",         OPT_STRING, &cmdlineP->arg0,
            &arg0Spec,                   0);
    OPTENT3(0, "nogroups",     OPT_FLAG,   NULL,
            &cmdlineP->nogroups,         0);
    OPTENT3(0, "username",     OPT_STRING, &cmdlineP->username,
            &usernameSpec,               0);
    OPTENT3(0, "asnobody",     OPT_FLAG,   NULL,
            &cmdlineP->asnobody,         0);
    
    opt.opt_table = option_def;
    opt.short_allowed = FALSE;  // We have no short (old-fashioned) options 
    opt.allowNegNum = FALSE;  // We have no parms that are negative numbers
    argc2 = argc;

    optParseOptions3(&argc2, argv, opt, sizeof(opt), 0);
        /* Uses and sets argc, argv, and some of *cmdlineP and others. */

    *errorP = NULL; // initial assumption

    if (!arg0Spec)
        cmdlineP->arg0 = NULL;

    specificUidOpts = ruidSpec + euidSpec + suidSpec + fsuidSpec;
    specificGidOpts = rgidSpec + egidSpec + sgidSpec + fsgidSpec;
    specificIdOpts  = specificUidOpts + specificGidOpts;

    if (cmdlineP->asnobody && specificIdOpts > 0)
        casprintf(errorP, "You can't specify -asnobody and also one of the "
                  "uid/gid options.");
    else if (usernameSpec && specificIdOpts > 0)
        casprintf(errorP, "You can't specify -username and also one of the "
                  "uid/gid options");
    else if (usernameSpec && cmdlineP->nogroups)
        casprintf(errorP, "You can't specify -username and -nogroups "
                  "together.");
    else {
        if (cmdlineP->asnobody) {
            makeNobodyIds(&cmdlineP->uidSet, &cmdlineP->gidSet);
            ++cmdlineP->nogroups;
        } else {
            if (!usernameSpec)
                cmdlineP->username = NULL;
            if (!ruidSpec)
                cmdlineP->uidSet.ruid  = -1;
            if (!euidSpec)
                cmdlineP->uidSet.euid  = -1;
            if (!suidSpec)
                cmdlineP->uidSet.suid  = -1;
            if (!fsuidSpec)
                cmdlineP->uidSet.fsuid = -1;
            if (!rgidSpec)
                cmdlineP->gidSet.rgid  = -1;
            if (!egidSpec)
                cmdlineP->gidSet.egid  = -1;
            if (!sgidSpec)
                cmdlineP->gidSet.sgid  = -1;
            if (!fsgidSpec)
                cmdlineP->gidSet.fsgid = -1;
        }
    }
    if (!*errorP) {
        if (cmdlineP->setCapabilities)
            parseCapabilityString(capabilities, 
                                  &cmdlineP->capabilities, errorP);
    }
    if (!*errorP) {
        if (argc2-1 < 1)
            casprintf(errorP, "You must specify the file name of the program "
                      "to exec as an argument.");
        else {
            cmdlineP->execProgName = argv[1];

            if (argc2-1 < 2) {
                static const char * emptyArgv[] = {NULL};
                cmdlineP->mainArgs = emptyArgv;
            } else
                cmdlineP->mainArgs = (const char **)&argv[2];
        }
    }
}



static void
setKeepcapsMode(void) {

    /*This prctl() disables the automatic clearing of permitted
      capabilities that happens by default when you switch away from
      euid or uid 0.  But it does not stop the clearing of effective
      capabilities.
    */
    prctl(PR_SET_KEEPCAPS, 1);
}    



static void
getMyCapabilities(processCaps * const existingCapsP,
                  const char ** const errorP) {


    getsetcapHeader header;
    int rc;

    header.version = _LINUX_CAPABILITY_VERSION;
    header.pid = 0;  // me

    rc = capget(&header, existingCapsP);

    if (rc != 0)
        casprintf(errorP, "capset() failed with errno %d (%s)", 
                  errno, strerror(errno));
    else {
        *errorP = NULL;
    }
}



static void
analyzeEpermFailure(processCaps   const newCaps,
                    const char ** const reasonP) {
    
    processCaps existingCaps;
    const char * error;

    getMyCapabilities(&existingCaps, &error);
    if (error) {
        casprintf(reasonP, "(Unable to analyze EPERM failure because attempt "
                  "to get current capabilities failed.  %s)", error);
        freeString(error);
    } else {
        if (!capsetIsSubset(newCaps.inheritable, existingCaps.permitted))
            casprintf(reasonP, 
                      "You requested capabilities that are not among your "
                      "permitted capabilities, so that is probably why.");
        else
            casprintf(reasonP, 
                      "Reason unknown.  It is not because you requested "
                      "any capability that is not among your permitted "
                      "capabilities");
    }
}



static void
setMyCapabilities(processCaps   const newCaps,
                  const char ** const errorP) {

    getsetcapHeader header;
    int rc;

    header.version = _LINUX_CAPABILITY_VERSION;
    header.pid = 0;  // me

    rc = capset(&header, &newCaps);
    if (rc != 0) {
        switch (errno) {
        case EPERM: {
            const char * epermReason;
            analyzeEpermFailure(newCaps, &epermReason);
                
            casprintf(errorP, "capset() failed with EPERM.  %s",
                      epermReason);
            freeString(epermReason);
        }
        break;
        default:
            casprintf(errorP, "capset() failed with errno %d (%s)\n",
                      errno, strerror(errno));
        }
    } else
        *errorP = NULL;
}



static void
setMyInhCapabilities(capset_t      const newInheritable,
                     const char ** const errorP) {
        
    processCaps existingCaps;
    const char * error;

    getMyCapabilities(&existingCaps, &error);

    if (error) {
        casprintf(errorP, "Unable to get the existing capabilities.  %s",
                  error);
        freeString(error);
    } else {
        processCaps newCaps;

        /* Our permitted and effective capability sets don't affect what
           capabilities we will have after exec.  Only our inheritable
           capabilities and capabilities of the file matter.  However, we
           are not permitted to set our inheritable capabilities to a 
           superset of our permitted capabilities, so have to make sure
           our permitted capabilites are sufficient.
        */
        newCaps = existingCaps;
        newCaps.inheritable = newInheritable;
        
        setMyCapabilities(newCaps, errorP);
    }
}



static void
setupAsUser(const char          username[],
            const char ** const errorP) {
/*----------------------------------------------------------------------------
   Set up the process ownership as if user username[] logged in.
-----------------------------------------------------------------------------*/
    struct passwd *passwdEntryP;

    passwdEntryP = getpwnam(username);

    if (passwdEntryP == NULL) 
        casprintf(errorP, "No such username '%s' (according to getpwnam())",
                  username);
    else {
        int rc;
        rc = setresgid(passwdEntryP->pw_gid, passwdEntryP->pw_gid,
                       passwdEntryP->pw_gid);
        if (rc != 0)
            casprintf(errorP, "Unable to set gids to %d.  "
                      "setresgid() fails with errno=%d (%s)", 
                      (int)passwdEntryP->pw_gid, errno, strerror(errno));
        else {
            int rc;
            rc = initgroups(username, passwdEntryP->pw_gid);
            if (rc != 0)
                casprintf(errorP, "Unable to set supplemental group list.  "
                          "initgroups() fails with errno %d (%s)",
                          errno, strerror(errno));
            else {
                int rc;
                rc = setresuid(passwdEntryP->pw_uid, passwdEntryP->pw_uid,
                               passwdEntryP->pw_uid);
                if (rc != 0)
                    casprintf(errorP, "Unable to set uids to %d.  "
                              "setresuid() fails with errno=%d (%s)", 
                              (int)passwdEntryP->pw_gid, 
                              errno, strerror(errno));
                else
                    *errorP = NULL;
            }
        }
    }
}



static void
setUidSet(struct uidSet const uidSet,
          const char ** const errorP) {

    processCaps capabilities;
    const char * error;
    
    getMyCapabilities(&capabilities, &error);
    if (error) {
        casprintf(errorP, "Unable to extract capabilities "
                  "so that they may be restored after setting uids.  %s",
                  error);
        freeString(error);
    } else {
        int rc;

        rc = setresuid(uidSet.ruid, uidSet.euid, uidSet.suid);
        
        if (rc != 0)
            casprintf(errorP, "Unable to set real, effective, and or "
                      "saved uids (%d, %d, %d).  "
                      "setresuid() failed with errno %d (%s)",
                      uidSet.ruid, uidSet.euid, uidSet.suid, 
                      errno, strerror(errno));
        else {
            const char * error;
            setMyCapabilities(capabilities, &error);
            if (error) {
                casprintf(errorP, "Unable to restore capabilities "
                          "after setting uids.  %s", error);
                freeString(error);
            } else {
                if (uidSet.fsuid != (uid_t)-1)
                    setfsuid(uidSet.fsuid);
                *errorP = NULL;
            }
        }
    }
}



static void
setGidSet(struct gidSet const gidSet,
          const char ** const errorP) {

    int rc;

    rc = setresgid(gidSet.rgid, gidSet.egid, gidSet.sgid);
    
    if (rc != 0)
        casprintf(errorP, "Unable to set real, effective, and/or saved gids "
                  "(%d, %d, %d).  "
                  "setresgid() failed with errno %d (%s)",
                  gidSet.rgid, gidSet.egid, gidSet.sgid, 
                  errno, strerror(errno));
    else {
        if (gidSet.fsgid != (gid_t)-1)
            setfsgid(gidSet.fsgid);
        *errorP = NULL;
    }
}



static void
setupExplicitId(struct cmdlineInfo const cmdline,
                const char **      const errorP) {
/*----------------------------------------------------------------------------
   Set up the uids, gid, and supplemental group list as explicitly
   described by the user.
-----------------------------------------------------------------------------*/
    if (!*errorP && cmdline.nogroups) {
        int rc;
        
        rc = setgroups(0, NULL);
        if (rc != 0)
            casprintf(errorP, "Unable to clear supplemental group list.  "
                      "setgroups() failed with errno %d (%s)",
                      errno, strerror(errno));
    }
    
    if (!*errorP)
        setGidSet(cmdline.gidSet, errorP);
    
    if (!*errorP)
        setUidSet(cmdline.uidSet, errorP);
}



static void
setupExec(struct cmdlineInfo const cmdline,
          const char **      const errorP) {

    const char * error;

    *errorP = NULL;  // initial value 

    if (!*errorP && cmdline.setCapabilities) {
        setMyInhCapabilities(cmdline.capabilities, &error);
            
        if (error)
            casprintf(errorP, "Unable to set inheritable capabilities.  %s",
                      error);
    }
    
    if (cmdline.username)
        setupAsUser(cmdline.username, errorP);
    else
        setupExplicitId(cmdline, errorP);
}



static const char *
lastComponent(const char * const fileName) {
/*----------------------------------------------------------------------------
   Return the last component of the file name 'fileName' in newly 
   malloc'ed storage.
-----------------------------------------------------------------------------*/
    const char * lastSlashPos;
    const char * p;

    lastSlashPos = NULL;
    p = &fileName[0];

    while (*p) {
        if (*p == '/')
            lastSlashPos = p;
        ++p;
    }
    if (lastSlashPos)
        return strdup(lastSlashPos + 1);
    else
        return strdup(fileName);
}



static void
computeExecArgs(const char *** const execArgvP,
                const char **  const mainArgs,
                const char *   const explicitArg0,
                const char *   const progName) {
    
    unsigned int mainArgsCount;
    const char ** execArgv;
    unsigned int argCount;
    const char * arg0;

    mainArgsCount = 0;
    while (mainArgs[mainArgsCount])
        mainArgsCount++;
        
    MALLOCARRAY_NOFAIL(execArgv, mainArgsCount + 1 + 1);
    
    argCount = 0;

    if (explicitArg0)
        arg0 = strdup(explicitArg0);
    else
        arg0 = lastComponent(progName);

    if (arg0 == NULL)
        abort();

    execArgv[argCount++] = arg0;
    
    {
        /* Copy mainArgs into the end of execArgv */
        unsigned int i;
        i = 0;
        while ((execArgv[argCount++] = mainArgs[i++]));
    }
    *execArgvP = execArgv;
}



int
main(int argc, char **argv) {

    struct cmdlineInfo cmdline;
    int retcode;
    const char * error;

    parseCommandLine(argc, argv, &cmdline, &error);
    if (error) {
        fprintf(stderr, "Invalid command.  %s\n", error);
        freeString(error);
        retcode = 100;
    } else {
        const char ** execArgv;
        const char * error;

        setKeepcapsMode();

        setupExec(cmdline, &error);
        if (error) {
            fprintf(stderr, "Unable to set up for exec.  %s\n", error);
            retcode = 40;
        } else {
            computeExecArgs(&execArgv, cmdline.mainArgs, cmdline.arg0,
                            cmdline.execProgName);

            execvp(cmdline.execProgName, (char**)execArgv);
            
            /* execv() failed.  (It doesn't return when it succeeds) */
            
            if (errno > 127) 
                retcode = 128;
            else 
                retcode = errno;
            
            fprintf(stderr, "Unable to execv() "
                    "program %s.  Errno = %d (%s)\n",
                    cmdline.execProgName, errno, strerror(errno));

            free(execArgv);
        }
    }
    return retcode;
}
