#define _GNU_SOURCE

#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/errno.h>
#include <linux/capability.h>

#include "giraffe/girtypes.h"
#include "giraffe/girstring.h"
#include "giraffe/mallocvar.h"
#include "shhopt/shhopt.h"

#include "gircap.h"


enum action {ACTION_SET, ACTION_ADD, ACTION_REMOVE};

struct cmdlineInfo {
    /* All the information the user supplied in the command line,
       in a form easy for the program to use.
    */
    pid_t    pid;
    capset_t capabilities;
    enum action action;
};



static void
parseCommandLine(int                  const argc, 
                 char **              const argv,
                 struct cmdlineInfo * const cmdlineP,
                 const char **        const errorP) {
/*----------------------------------------------------------------------------
   Parse program command line described in Unix standard form by argc
   and argv.  Return the information in the options as *cmdlineP.  

   If command line is internally inconsistent (invalid options, etc.),
   issue error message to stderr and abort program.

   Note that the strings we return are stored in the storage that
   was passed to us as the argv array.  We also trash *argv.
-----------------------------------------------------------------------------*/
    optEntry *option_def;
        /* Instructions to optParseOptions3 on how to parse our options.
         */
    optStruct3 opt;
    int argc2;

    unsigned int option_def_index;

    char * capabilities;
    unsigned int pidSpec, capabilitiesSpec;
    unsigned int set, add, remove;

    MALLOCARRAY_NOFAIL(option_def, 100);

    option_def_index = 0;   /* incremented by OPTENT3 */
    OPTENT3(0, "pid",          OPT_UINT,   &cmdlineP->pid,
            &pidSpec,                    0);
    OPTENT3(0, "capabilities", OPT_STRING, &capabilities,
            &capabilitiesSpec,           0);
    OPTENT3(0, "set",          OPT_FLAG, NULL,
            &set,                        0);
    OPTENT3(0, "add",          OPT_FLAG, NULL,
            &add,                        0);
    OPTENT3(0, "remove",       OPT_FLAG, NULL,
            &remove,                     0);

    opt.opt_table = option_def;
    opt.short_allowed = FALSE;  // We have no short (old-fashioned) options 
    opt.allowNegNum = FALSE;  // We have no parms that are negative numbers
    argc2 = argc;

    optParseOptions3(&argc2, argv, opt, sizeof(opt), 0);
        /* Uses and sets argc, argv, and some of *cmdlineP and others. */

    *errorP = NULL; // initial assumption
    
    if (!*errorP) {
        if (!pidSpec)
            casprintf(errorP, "You must specify the -pid option.");
    }
    if (!*errorP) {
        switch (set + add + remove) {
        case 0:
            cmdlineP->action = ACTION_SET;
            break;
        case 1:
            cmdlineP->action = 
                set ? ACTION_SET : add ? ACTION_ADD : ACTION_REMOVE;
            break;
        default:
            casprintf(errorP, "You can specify only one of "
                      "-set, -add, and -remove");
        }
    }
    if (!*errorP) {
        if (capabilitiesSpec)
            parseCapabilityString(capabilities, 
                                  &cmdlineP->capabilities, errorP);
    }
    if (!*errorP) {
        if (argc2-1 != 0)
            casprintf(errorP, 
                      "There are no non-option arguments.  You specifed %d", 
                      argc2-1);
    }
}



static void
analyzeEpermFailure(struct __user_cap_data_struct const newCaps,
                    pid_t                         const pid,
                    const char **                 const reasonP) {

    struct __user_cap_data_struct myCaps;
    struct __user_cap_header_struct header;
    int rc;
    
    header.version = _LINUX_CAPABILITY_VERSION;
    header.pid = 0;  // me

    rc = capget(&header, &myCaps);

    if (rc != 0)
        casprintf(reasonP, "We could not "
                  "investigate why, because capset() of self also failed.");
    else {
        if (!(myCaps.effective & CAP_SETPCAP))
            casprintf(reasonP, "You do not "
                      "have SETPCAP capability, so that is probably why.");
        else {
            int rc;

            struct __user_cap_header_struct header;
            struct __user_cap_data_struct oldCaps;

            header.version = _LINUX_CAPABILITY_VERSION;
            header.pid = pid;

            rc = capget(&header, &oldCaps);
            
            if (rc != 0)
                casprintf(reasonP, "We could not investigate why because "
                          "capget() on pid %d also failed.", pid);
            else {
                if (!capsetIsSubset(newCaps.inheritable,
                                    capsetUnion(oldCaps.inheritable,
                                                myCaps.permitted)))
                    casprintf(reasonP, "You requested to add new capabilities "
                              "to the inheritable set that are not among "
                              "your own permitted capabilities, so that is "
                              "probably why.");
                else if (!capsetIsSubset(newCaps.permitted,
                                         capsetUnion(oldCaps.permitted,
                                                     myCaps.permitted)))
                    casprintf(reasonP, "You requested to add new capabilities "
                              "to the permitted set that are not among "
                              "your own permitted capabilities, so that is "
                              "probably why.");
                else if (!capsetIsSubset(newCaps.effective, newCaps.permitted))
                    casprintf(reasonP, "You requested to make the effective "
                              "capabilities something that is not a subset "
                              "of what you requested for the permitted set, "
                              "so that is probably why.");
                else
                    casprintf(reasonP, "I can't tell why.  It is not that "
                              "you lack SETPCAP capability or that you're "
                              "trying to give the process capabilities you "
                              "don't possess or that you're trying to make "
                              "the effective capabilities not a subset of "
                              "the permitted capabilities.");
            }
        }
    }
}



static void
setCapabilities(pid_t         const pid,
                processCaps   const requestedCapabilities,
                const char ** const errorP) {
        
    struct __user_cap_data_struct newCaps;
    struct __user_cap_header_struct header;
    int rc;

    header.version = _LINUX_CAPABILITY_VERSION;
    header.pid = pid;

    newCaps.permitted   = requestedCapabilities.permitted;
    newCaps.effective   = requestedCapabilities.effective;
    newCaps.inheritable = requestedCapabilities.inheritable;

    rc = capset(&header, &newCaps);
    if (rc != 0) {
        const char * reason;
        switch (errno) {
        case EPERM: {
            const char * epermReason;
            analyzeEpermFailure(newCaps, pid, &epermReason);

            casprintf(&reason, "capset() failed with EPERM.  %s",
                      epermReason);
            freeString(epermReason);
        }
        break;
        default:
            casprintf(&reason, "capset() failed with errno %d (%s)\n",
                      errno, strerror(errno));
        }
        casprintf(errorP, "Unable to set capabilities.  %s", reason);
        freeString(reason);
    }
    else
        *errorP = NULL;
}




static void
getCapabilities(pid_t         const pid,
                processCaps * const currentCapsP,
                const char ** const errorP) {

    struct __user_cap_data_struct myCaps;
    struct __user_cap_header_struct header;
    int rc;
    
    header.version = _LINUX_CAPABILITY_VERSION;
    header.pid = pid;

    rc = capget(&header, &myCaps);

    if (rc != 0)
        casprintf(errorP, "capset() failed.  errno=%d (%s)",
                  errno, strerror(errno));
    else {
        *errorP = NULL;

        currentCapsP->permitted   = myCaps.permitted;
        currentCapsP->effective   = myCaps.effective;
        currentCapsP->inheritable = myCaps.inheritable;
    }
}



static processCaps
addCaps(processCaps const currentCaps,
        capset_t    const addedCaps) {

    processCaps newCapabilities;

    newCapabilities.permitted =
        capsetUnion(currentCaps.permitted,   addedCaps);
    newCapabilities.effective =
        capsetUnion(currentCaps.effective,   addedCaps);
    newCapabilities.inheritable =
        capsetUnion(currentCaps.inheritable, addedCaps);

    return newCapabilities;
}
    


static processCaps
delCaps(processCaps const currentCaps,
        capset_t    const addedCaps) {

    processCaps newCapabilities;

    newCapabilities.permitted =
        capsetIntersection(currentCaps.permitted,   
                           capsetInverse(addedCaps));
    newCapabilities.effective =
        capsetIntersection(currentCaps.effective,   
                           capsetInverse(addedCaps));
    newCapabilities.inheritable =
        capsetIntersection(currentCaps.inheritable, 
                           capsetInverse(addedCaps));

    return newCapabilities;
}
    


static void
computeNewCapabilities(struct cmdlineInfo const cmdline,
                       processCaps *      const newCapabilitiesP,
                       const char **      const errorP) {

    if (cmdline.action == ACTION_SET) {
        newCapabilitiesP->permitted   = cmdline.capabilities;
        newCapabilitiesP->effective   = cmdline.capabilities;
        newCapabilitiesP->inheritable = cmdline.capabilities;
        *errorP = NULL;
    } else {
        /* It's relative, so we have to get the old capabilities */

        processCaps currentCaps;
        const char * error;

        getCapabilities(cmdline.pid, &currentCaps, &error);

        if (error) {
            casprintf(errorP, "Could not query the existing capabilities.  %s",
                      error);
            freeString(error);
        } else {
            *errorP = NULL;
            switch (cmdline.action) {
            case ACTION_ADD:
                *newCapabilitiesP = addCaps(currentCaps, cmdline.capabilities);
                break;
            case ACTION_REMOVE:
                *newCapabilitiesP = delCaps(currentCaps, cmdline.capabilities);
                break;
            case ACTION_SET:
                fprintf(stderr, "INTERNAL ERROR: action is ACTION_SET where "
                        "it shouldn't be.\n");
                abort();
                break;
            }
        }
    }
}



int
main(int argc, char **argv) {

    struct cmdlineInfo cmdline;
    const char * error;
    int retcode;

    parseCommandLine(argc, argv, &cmdline, &error);
    if (error) {
        fprintf(stderr, "Invalid options.  %s\n", error);
        freeString(error);
        retcode = 100;
    } else {
        processCaps newCapabilities;
        computeNewCapabilities(cmdline, &newCapabilities, &error);
        if (error) {
            fprintf(stderr, "Unable to compute the capabilities to set.  %s",
                    error);
            freeString(error);
            retcode = 1;
        } else {
            setCapabilities(cmdline.pid, newCapabilities, &error);

            if (error) {
                fprintf(stderr, "Failed.  %s\n", error);
                freeString(error);
                retcode = 2;
            } else
                retcode = 0;
        }
    }
    return retcode;
}

