#ifndef GIRCAP_INCLUDED
#define GIRCAP_INCLUDED

#include <linux/capability.h>

#include <giraffe/girstring.h>

typedef __u32 capset_t;
static const capset_t noCaps = 0u;
typedef unsigned int capability_t;
    /* A 32 bit mask with one bit set */
typedef struct __user_cap_data_struct processCaps;
typedef struct __user_cap_header_struct getsetcapHeader;


static inline capset_t
capsetInverse(capset_t const capset) {

    return ~capset;
}



static inline capset_t
capsetUnion(capset_t const capsetA,
            capset_t const capsetB) {

    return capsetA | capsetB;

}



static inline capset_t
capsetIntersection(capset_t const capsetA,
                   capset_t const capsetB) {

    return capsetA & capsetB;
}



static inline bool
capsetIsSubset(capset_t const subset,
               capset_t const superset) {

    return (subset & superset) == subset;
}



static void
addCapability(capset_t *   const capsetP,
              capability_t const capability) {

    *capsetP |= capability;
}


int capget(const getsetcapHeader * const headerP, 
           processCaps *           const dataP);
int capset(const getsetcapHeader * const headerP,
           const processCaps *     const dataP);


static struct {
    const char * name;
    capability_t capability;
} capTable[] = {
    {"chown",            1 << CAP_CHOWN},
    {"dac_override",     1 << CAP_DAC_OVERRIDE},
    {"dac_read_search",  1 << CAP_DAC_READ_SEARCH},
    {"fowner",           1 << CAP_FOWNER},
    {"fsetid",           1 << CAP_FSETID},
    {"kill",             1 << CAP_KILL},
    {"setgid",           1 << CAP_SETGID},
    {"setuid",           1 << CAP_SETUID},
    {"setpcap",          1 << CAP_SETPCAP},
    {"linux_immutable",  1 << CAP_LINUX_IMMUTABLE},
    {"net_bind_service", 1 << CAP_NET_BIND_SERVICE},
    {"net_broadcast",    1 << CAP_NET_BROADCAST},
    {"net_admin",        1 << CAP_NET_ADMIN},
    {"net_raw",          1 << CAP_NET_RAW},
    {"ipc_lock",         1 << CAP_IPC_LOCK},
    {"ipc_owner",        1 << CAP_IPC_OWNER},
    {"sys_module",       1 << CAP_SYS_MODULE},
    {"sys_rawio",        1 << CAP_SYS_RAWIO},
    {"sys_chroot",       1 << CAP_SYS_CHROOT},
    {"sys_ptrace",       1 << CAP_SYS_PTRACE},
    {"sys_pacct",        1 << CAP_SYS_PACCT},
    {"sys_admin",        1 << CAP_SYS_ADMIN},
    {"sys_boot",         1 << CAP_SYS_BOOT},
    {"sys_nice",         1 << CAP_SYS_NICE},
    {"sys_resource",     1 << CAP_SYS_RESOURCE},
    {"sys_time",         1 << CAP_SYS_TIME},
    {"sys_tty_config",   1 << CAP_SYS_TTY_CONFIG},
};



static inline void
parseCap(const char *  const value,
         capset_t *    const capsetP,
         const char ** const errorP) {

    unsigned int i;
    bool found;

    for (i = 0, found = FALSE; i < sizeof(capTable)/sizeof(capTable[0]); ++i) {
        if (strcmp(value, capTable[i].name) == 0) {
            addCapability(capsetP, capTable[i].capability);
            found = TRUE;
            break;
        }
    }
    if (!found)
        casprintf(errorP, "Unrecognized capability name '%s'", value);
    else
        *errorP = NULL;
}


static inline void
parseCapabilityString(const char *  const capabilityString,
                      capset_t *    const capabilitiesP,
                      const char ** const errorP) {
/*----------------------------------------------------------------------------
  Parse the value of a capability string (e.g. "chown,setuid,setgid") 
  into capabilities in the capset_t form.
-----------------------------------------------------------------------------*/
    char const delimiters[] = ",";

    capset_t capsSoFar;
    char * capWork;
    char *cursor;
    char *token;
    bool negative;

    capWork = strdupa(capabilityString);
    capsSoFar = noCaps;

    if (capabilityString[0] == '!') {
        negative = TRUE;
        cursor = &capWork[1];
    } else {
        negative = FALSE;
        cursor = &capWork[0];
    }

    if (*cursor == '\0')
        cursor = NULL;   // patch up a hole in strsep()

    while (!*errorP && (token = strsep(&cursor, delimiters))) {
        parseCap(token, &capsSoFar, errorP);
    }
    if (negative)
        *capabilitiesP = capsetInverse(capsSoFar);
    else
        *capabilitiesP = capsSoFar;
}



#endif
