/*----------------------------------------------------------------------------
                              binfmt_capx
-------------------------------------------------------------------------------
  This is a Linux binary handler.  It runs a "capx" file.  A capx file
  names an executable and describes capabilities and other process
  parameters with which it should run.  This handler sets up the process
  thus and then execs the named executable.

  The handler recognizes a capx file as one that starts with ASCII
  "CAPX" and is setuid and owned by uid 0.
-----------------------------------------------------------------------------*/

#define __KERNEL__
#define MODULE

#include <linux/module.h>
#include <linux/string.h>
#include <linux/stat.h>
#include <linux/slab.h>
#include <linux/binfmts.h>
#include <linux/init.h>
#include <linux/file.h>
#include <linux/smp_lock.h>
#include <asm/byteorder.h>
#include <linux/capability.h>


#define STRSCPY(to, from) { \
    strncpy(to, from, sizeof(to)); \
    to[sizeof(to)-1] = '\0';\
    }

struct capxControl {
    /* This is the contents of a CAPX file, in C format */

    char filenameToExec[64];
    struct {
        /* Masks to get applied to the capabilities which are naturally
           associated with the CAPX file before computing the capabilities
           that theprocess will have after exec.  These always subtract
           capabilities; they cannot add them.
        */
        kernel_cap_t permitted;
        kernel_cap_t effective;
        kernel_cap_t inheritable;
    } capabilityMask;
};



struct fileFormat {
    /* These integers aren't really C integers -- they're 4 byte
       big-endian binary integers or arrays of bits.  We declare them
       this way because it makes it easiest to convert them into
       integers, capability sets, etc.
    */
    unsigned int optionalFeatures;
    /* Flags that represent features that new interpreters can exploit
       but old interpreters can ignore.
    */
    unsigned int mandatoryFeatures;
    /* Flags that represent features that all interpreters have to
       deal with.  Old interpreters that don't know how must fail the
       exec.  
    */
    unsigned int permittedMask;
    unsigned int effectiveMask;
    unsigned int inheritableMask;
    unsigned char filenameToExec[64];
};



static void
decodeCapxControl(unsigned char *      const fileContents,
                  unsigned int         const fileContentsSize,
                  struct capxControl * const capxControlP,
                  int *                const errorP) {
/*----------------------------------------------------------------------------
   Decode the contents of the file, starting right after the 4 characters
   "CAPX".

   *fileContents is the contents of the file after the part that indicates
   that it is a CAPX file.  'fileContentsSize' is the storage size of the
   data structure, but we have no indication of how much data is actually
   in the file.  If the file is big, the contents are truncated.  If it
   is small, *fileContents is filled with zeroes.
-----------------------------------------------------------------------------*/
    struct fileFormat * const contentsP = (struct fileFormat *)fileContents;

    if (sizeof(*contentsP) > fileContentsSize)
        *errorP = ENOEXEC;
    else {
        if (be32_to_cpu(contentsP->mandatoryFeatures) != 0)
            
            *errorP = ENOEXEC;
        else {
            STRSCPY(capxControlP->filenameToExec, contentsP->filenameToExec);

            capxControlP->capabilityMask.permitted = 
                to_cap_t(be32_to_cpu(contentsP->permittedMask));
            capxControlP->capabilityMask.effective = 
                to_cap_t(be32_to_cpu(contentsP->effectiveMask));
            capxControlP->capabilityMask.inheritable = 
                to_cap_t(be32_to_cpu(contentsP->inheritableMask));
            *errorP = 0;
        }
    }
}



static void
maskCapabilities(struct linux_binprm * const bprmP,
                 struct capxControl    const capxControl) {
/*----------------------------------------------------------------------------
   Reduce the 3 capability sets of the file, in *bprmP, by the masks
   supplied with it.

   In normal operation, the capabilities that are now in *bprmP were
   inferred from the uid and the setuid flag of the file.  

   - If the file is setuid 0, they are all full capabilities, so the
     capxControl masks simply take there places as the file's capability
     sets.

   - If the file is setuid nonzero, they are all zero capabilities, so
     the capxControl masks have no effect.

   - If the file is not setuid, the permitted and effective
     capabilities are zero, so remain zero, but the inheritable
     capabilities are full, so get replaced by the capxControl inheritable
     capability mask.
-----------------------------------------------------------------------------*/


    bprmP->cap_permitted = 
        cap_intersect(bprmP->cap_permitted, 
                      capxControl.capabilityMask.permitted);
    bprmP->cap_effective = 
        cap_intersect(bprmP->cap_effective, 
                      capxControl.capabilityMask.effective);
    bprmP->cap_inheritable = 
        cap_intersect(bprmP->cap_inheritable, 
                      capxControl.capabilityMask.inheritable);

}



static void
setFromFile(struct linux_binprm * const bprmP,
            int *                 const fileAccessRcP) {
/*----------------------------------------------------------------------------
   Set the members of *bprmP which are based on information in the file
   to be execed or its metadata.

   *fileAccessRcP is a negative errno for failure, number of bytes
   read from the file for success.
-----------------------------------------------------------------------------*/
    *fileAccessRcP = prepare_binprm(bprmP);

    /* prepare_binprm() sets bprmP->euid, bprmP->egid,
       bprmP->buf, bprmP->cap_effective, bprmP->cap_permitted,
       bprmP->cap_inheritable.
    */
}



static void
capxExec(struct linux_binprm * const bprmP,
         struct pt_regs *      const regsP,
         int *                 const retvalP) {
             
    int error;
    struct capxControl capxControl;
    
    decodeCapxControl(&bprmP->buf[4], BINPRM_BUF_SIZE-4,
                      &capxControl, &error);
    
    if (error)
        *retvalP = -error;
    else {
        struct file * filpP;
        
        maskCapabilities(bprmP, capxControl);

        compute_creds(bprmP); // Sets euid, egid, caps

        printk("CAPX running %s with perm=%08X\n", 
               capxControl.filenameToExec, current->cap_permitted);
        filpP = open_exec(capxControl.filenameToExec);
        if (IS_ERR(filpP))
            *retvalP = PTR_ERR(filpP);
        else {
            int fileAccessRc;
            bprmP->filename = capxControl.filenameToExec;
            bprmP->file = filpP;
            /* We leave args & environment the same */
            setFromFile(bprmP, &fileAccessRc);

            if (fileAccessRc < 0)
                *retvalP = fileAccessRc;
            else {
                *retvalP = search_binary_handler(bprmP, regsP);
                printk("search_binary_handler() returns %d\n", *retvalP);
            }
        }
    }
}



static int
stackTooDeep(void) {

#ifdef CONFIG_X86
    /* The task_struct is on an 8K boundary and the stack is above it
       and grows down toward it.  So we just take the stack pointer
       mod 8K and see if there's at least 1K of space between it and
       the upper end of the task_struct.
    */
    unsigned long esp;  /* Value of stack pointer register (ESP) */

    __asm__ __volatile__("movl %%esp,%0" : "=r" (esp));

    return (esp % 8192 - sizeof(struct task_struct) < 1024);
#else
    /* We don't know how to check the stack */
    return 0;
#endif
}



static int 
capxTryExec(struct linux_binprm * const bprmP, 
            struct pt_regs      * const regsP) {

    int retval;

    if (memcmp(&bprmP->buf[0], "CAPX", 4) != 0) {
        retval = -ENOEXEC;
    } else {
        printk("It's CAPX\n");

        allow_write_access(bprmP->file);
            /* Let people write to the file without ETXTBSY now */
        fput(bprmP->file);
        bprmP->file = NULL;

        /* Here is our check on infinite recursion -- e.g. a CAPX
           file says to execute itself.
        */

        if (stackTooDeep()) {
            /* ELOOP is a good hint, since this is similar to a symlink
               loop.
            */
            retval = -ELOOP;
        } else
            capxExec(bprmP, regsP, &retval);
    }
    return retval;
}



static struct linux_binfmt capxFormat = {
    NULL, THIS_MODULE, capxTryExec, NULL, NULL, 0
};

static int __init 
initializeBinfmtDriver(void) {
    return register_binfmt(&capxFormat);
}

static void __exit 
terminateBinfmtDriver(void) {
    unregister_binfmt(&capxFormat);
}

module_init(initializeBinfmtDriver);
module_exit(terminateBinfmtDriver);
